from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from core.base import base_likelihood_model
import torch
import math
import numpy as np


class Gaussian(base_likelihood_model.Likelihood):
    """
        Class for the Gaussian PDF.
        Methods:
            - logLikelihood
                inputs: signal (measured signal), mu (simulated signal) and sigma (SD of the noise)
                outputs: data consistency loss
            - applyNoise
                inputs: a signal and sigma
                outputs: Noisy signal corrupted by additive gaussian noise
    """
    
    def __init__(self, config_object):
        super(Gaussian, self).__init__(config_object, self)
        self.__name__ = 'Gaussian'
        self.args = config_object.args

    def likelihood(self, signal, modeled_signal):
        return torch.sum((signal - modeled_signal)**2)

    def apply_noise(self, signal, sigma):
        signal += torch.from_numpy(np.random.normal(0.0, sigma, signal.size())).to(self.args.engine.device)
        return signal

    


from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
import torch
import logging


def save(args, epoch, model):
    """
        Save trained DL model to checkpoint.
        Inputs: args (console inputs), epoch and model

    """
    path = os.path.join(args.engine.saveCheckpointPath, args.engine.inferenceModel+'_epoch_'+ str(epoch) + '_' + args.engine.sulfixCheckpoint + '.pth')
    torch.save({'epoch': epoch,
                'model_state': model.state_dict(),
                'optimizer_state': args.engine.optimizer.state_dict()}, path)

    logging.info('Model saved to: {}'.format(path))


def load(config_object):
    """
        Load saved DL model from checkpoint.
        Inputs: args (console inputs), epoch and model
        Outputs: DL model, optimiser checkpoint and epoch
    """
    try:
        checkpoint = torch.load(config_object.args.engine.loadCheckpointPath, map_location=config_object.args.engine.device)
        try:
            config_object.args.engine.inference_model.load_state_dict(checkpoint['model_state'])
            config_object.args.engine.optimizer.load_state_dict(checkpoint['optimizer_state'])
        except KeyError:
            try:
                config_object.args.engine.inference_model.load_state_dict(checkpoint['model_state_dict'])
                config_object.args.engine.optimizer.load_state_dict(checkpoint['optimizer_state_dict'])
            except KeyError:
                logging.error('Error importing pre-trained model')

        epoch = checkpoint['epoch']

    except AssertionError:
        logging.error("Checkpoint path not specified. Please insert a valid checkpoint path")
        return -1



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from torch.utils.data import DataLoader
from torch.utils.data import Dataset
from pathlib import Path
from ..utilities import core_utilities

import numpy as np

class SampleDatabase(Dataset):
    """ Wrapper function for custom dataset subclasses.
    """
    def __init__(self, config_object, path=[]):
        self.args = config_object.args
        self.dataset = config_object.args.engine.dataset_model
        self.dataset.set_data_path(path)
        self.dataset.index_files()

    def __len__(self):
        return self.dataset.get_length()

    def __getitem__(self, idx):
        self.signal = self.dataset.get_signal(idx)
        self.label, self.mask = self.dataset.get_label(idx)
        self.filename = Path(self.dataset.filename_list[idx]).stem
        self.extension = Path(self.dataset.filename_list[idx]).suffix
        
        data = {}
        data['filename'] = self.filename
        data['file_extension'] = self.extension
        data['image'] = self.signal
        if self.label is None:
            data['label'] = []
            data['mask'] = []
        else:
            data['label'] = self.label
            data['mask'] = self.mask

        return data



def create_dataloader(config_object):
    """Creates the dataloader handler

    Args:
        config_object ([Configuration]): [Object containing all backend configuration settings]
        Required config_object.args:
            - mode
            - dataset
            - trainingDataPath, validationDataPath and/or testingDataPath
            - batchSize
            - usePatches
            - useSimulatedData
            - runValidation
            
    Returns:
        [type: List]: List of DataLoader objects. If mode=='training', it returns 
                      [trainingDataloader, validationDataloader]. If mode=='testing', it returns 
                      [testingDataloader]
    """
    if config_object.args.engine.mode == 'training':
        training_path = Path(config_object.args.engine.trainingDataPath).absolute()
        validation_path = Path(config_object.args.engine.validationDataPath).absolute()

        training_database_handler = SampleDatabase(config_object, path=training_path)
        trainingDataloader = DataLoader(training_database_handler,
                                batch_size = config_object.args.engine.batchSize,
                                shuffle = config_object.args.engine.shuffleData,
                                num_workers = config_object.args.engine.numWorkers
                                )
        if config_object.args.engine.runValidation:
            validation_database_handler = SampleDatabase(config_object, path=validation_path)
            validationDataloader = DataLoader(validation_database_handler,
                                    batch_size = config_object.args.engine.batchSize,
                                    shuffle = config_object.args.engine.shuffleData,
                                    num_workers = config_object.args.engine.numWorkers
                                    )
            dataloader = trainingDataloader, validationDataloader
        else:
            dataloader = trainingDataloader

    elif config_object.args.engine.mode == 'testing':
        testing_path = Path(config_object.args.engine.testingDataPath).absolute()
        testing_database_handler = SampleDatabase(config_object, path=testing_path)
        testingDataloader = DataLoader(testing_database_handler,
                                batch_size = config_object.args.engine.batchSize,
                                shuffle = config_object.args.engine.shuffleData,
                                num_workers = config_object.args.engine.numWorkers
                                )
        dataloader = testingDataloader

    config_object.args.engine.dataloader = dataloader


from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import sys
sys.path.insert(0, '../')

from ..utilities import image_utilities
import logging
import torch
import torch.nn as nn

logging.basicConfig(level=logging.INFO)
logging.info('Setting up environment...')

class Infer(object):
    """Performs a forward pass through a given inference model.
        Depending on the options set, it might save intermediate results and training checkpoints
        and/or display the intermediate results.

    Args:
        config_object ([type: Configuration]): [Object containing all backend configuration settings]
        Required config_object.args:
            - epochs
            - inference_model
            - dataloader
            - signal_model (if config_object.args.inference_model.__require_initial_guess__ == True)
            - numberOfPatches
            - objective_fun
            - optimizer
            - saveResults
            - saveResultsPath
            - saveCheckpoint
            - saveCheckpointPath
            - usePatchesAsBatches
    """

    def __init__(self, device, validation_dataloader, network, metric, prepare_batch, compute_loss, log_training_fun, config_object):
        self.args = config_object.args
        self.device = device
        self.dataloader = validation_dataloader
        self.network = network
        self.metric = metric
        self.prepare_batch = prepare_batch
        self.compute_loss = compute_loss
        self.log_training_fun = log_training_fun

    def run(self, return_result = False):
        self.args.engine.batch_size_iter = 1
        if isinstance(self.network, nn.Module):
            self.network.eval()

        epoch = '_estimating_'
        logging.info('Running {}'.format(self.args.engine.state_name))
        logging.info('*'*50)
        self.args.engine.len_dataset = len(self.dataloader)
        for i, data in enumerate(self.dataloader):
            self.args.engine.iter = i
            data_ = self.prepare_batch(data, self.device)
            processed_data = self.__run__(data_, i)
            if self.log_training_fun:
                self.log_training_fun(processed_data, 0, 1, 1)

            if self.args.engine.saveResults:
                self.args.engine.filename = data['filename'][0]
                filename_ = 'estimated_parameters_' + self.args.engine.filename 
                image_utilities.saveData(processed_data, self.args.engine.saveResultsPath, filename_, data['file_extension'][0])
                
        if return_result:
            return processed_data

    def __log_error__(self, i, loss):
        logging.info("Sample: {}/{}, Loss: {} ".format(i+1, self.dataloader.__len__(), loss))

    def __run__(self, data_, i):
        signal = data_['image']
        label = data_['label'] if len(data_['label']) else []

        estimate = self.network.forward(signal)

        if self.compute_loss:
            if label is not None:
                loss = self.args.engine.objective_fun(estimate, label.expand_as(estimate))
                self.__log_error__(i, loss)
            else:
                logging.error("No label provided")
        
        processed_data = {}
        processed_data['estimated'] = estimate
        processed_data['signal'] = signal
        if label is not None:
            processed_data['label'] = label

        return processed_data



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import importlib
import logging
import progress.bar
import torch


def prep_batch(batchdata, device, non_blocking=False):
    batchdata['image']=batchdata['image'].to(device=device, non_blocking=non_blocking)
    if batchdata['label']:
        batchdata['label']=batchdata['label'].to(device=device, non_blocking=non_blocking)
    return batchdata
    

def get_engine(config_object):

    if not hasattr(config_object.args.engine, 'prepare_batch'):
        config_object.args.engine.prepare_batch = prep_batch
    
    if config_object.args.engine.state_name == 'training':
        if config_object.args.engine.trainerModule == 'monai':
            from monai.engines import SupervisedTrainer
            from monai.handlers import StatsHandler
            trainer_ = SupervisedTrainer(device=config_object.args.engine.device,
                        max_epochs=config_object.args.engine.epochs,
                        train_data_loader=config_object.args.engine.dataloader,
                        network=config_object.args.engine.inference_model,
                        optimizer=config_object.args.engine.optimizer,
                        loss_function=config_object.args.engine.objective_fun,
                        train_handlers=StatsHandler(tag_name="train_loss", output_transform=lambda x: x["loss"]),
                        prepare_batch=config_object.args.engine.prepare_batch
                )
            config_object.args.engine.trainer = trainer_
        elif config_object.args.engine.trainerModule == 'emcqmri':
            from core.engine.train_model import Trainer
            trainer_ = Trainer(device = config_object.args.engine.device,
                            max_epochs=config_object.args.engine.epochs,
                            train_data_loader=config_object.args.engine.dataloader,
                            network=config_object.args.engine.inference_model,
                            optimizer=config_object.args.engine.optimizer,
                            loss_function=config_object.args.engine.objective_fun,
                            prepare_batch = config_object.args.engine.prepare_batch,
                            log_training_fun=config_object.args.engine.log_training_fun,
                            config_object = config_object
                )
            config_object.args.engine.trainer = trainer_
        else:
            logging.error("Selected trainer not available")
            
    elif config_object.args.engine.state_name == 'testing':
        if config_object.args.engine.estimatorModule == 'emcqmri':
            from core.engine.estimate import Infer
            estimator_ = Infer(device = config_object.args.engine.device,
                            validation_dataloader=config_object.args.engine.dataloader,
                            network=config_object.args.engine.inference_model,
                            metric=config_object.args.engine.objective_fun,
                            prepare_batch = config_object.args.engine.prepare_batch,
                            compute_loss = False,
                            log_training_fun=config_object.args.engine.log_training_fun,
                            config_object = config_object
                    )
            config_object.args.engine.estimator = estimator_
        else:
            logging.error("Only the 'emcqmri' estimator is available.")


def load_ext_module(module_link, module_name, config_object):
    module = importlib.import_module(module_link)
    loaded_module = str_to_class(module, module_name.capitalize())
    return loaded_module(config_object)


def str_to_class(module, classname):
    return getattr(module, classname)


class ProgressWrapper(progress.bar.FillingSquaresBar):
    def __init__(self, *args, **kwargs):
        super(ProgressWrapper, self).__init__(*args, **kwargs)
    
    def update(self):
        filled_length = int(self.width * self.progress)
        empty_length = self.width - filled_length
        message = self.message % self
        bar_ = self.fill * filled_length
        empty = self.empty_fill * empty_length
        suffix = self.suffix % self
        self.loss = '  Loss:  '+str(self.loss_)
        self.subs = 'Subject: '+str(self.it_sub)+'/'+str(self.total_subs) + '  '
        line = ''.join([self.subs, message, self.bar_prefix, bar_, empty, self.bar_suffix,
                        suffix, self.loss])
        self.writeln(line)

    def set_total_sub(self, total_subs):
        self.total_subs = total_subs

    def set_max(self, max_):
        self.max = max_

    def reset(self):
        self.index = 0
        self.loss_ = 0
        self.it_sub = 0

    def update_ext_par(self, loss, it_sub):
        self.loss_ = loss
        self.it_sub = it_sub


def ProgressBarWrap(func):
    bar = ProgressWrapper('', max=1, suffix='%(percent)d%%', loss=10)
    def wrapper(self, loss, args):
        bar.set_max(args.inference.inferenceSteps)
        bar.set_total_sub(args.engine.dataloader.__len__())
        bar.update_ext_par(loss, args.engine.iter)
        bar.next()
        
        if bar.index == args.inference.inferenceSteps:
            bar.finish()
            bar.reset()
    return wrapper





from torch import nn
from torch.nn import Module


def determine_conv_class(n_dim, transposed=False):
    if n_dim == 1:
        if not transposed:
            return nn.Conv1d
        else:
            return nn.ConvTranspose1d
    elif n_dim == 2:
        if not transposed:
            return nn.Conv2d
        else:
            return nn.ConvTranspose2d
    elif n_dim == 3:
        if not transposed:
            return nn.Conv3d
        else:
            return nn.ConvTranspose3d
    else:
        NotImplementedError("No convolution of this dimensionality implemented")


def determine_conv_functional(n_dim, transposed=False):
    if n_dim == 1:
        if not transposed:
            return nn.functional.conv1d
        else:
            return nn.functional.conv_transposed1d
    elif n_dim == 2:
        if not transposed:
            return nn.functional.conv2d
        else:
            return nn.functional.conv_transposed2d
    elif n_dim == 3:
        if not transposed:
            return nn.functional.conv3d
        else:
            return nn.functional.conv_transposed3d
    else:
        NotImplementedError("No convolution of this dimensionality implemented")


def pixel_unshuffle(x, downscale_factor):
    b, c, h, w = x.size()
    r = downscale_factor
    out_channel = c * (r ** 2)
    out_h = h // r
    out_w = w // r
    x_view = x.contiguous().view(b, c, out_h, r, out_w, r)
    x_prime = x_view.permute(0, 1, 3, 5, 2, 4).contiguous().view(b, out_channel, out_h, out_w)

    return x_prime


class PixelUnshuffle(Module):
    def __init__(self, downscale_factor):
        super(PixelUnshuffle, self).__init__()
        self.downscale_factor = downscale_factor

    def forward(self, x):
        return pixel_unshuffle(x, self.downscale_factor)



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from core.base import base_inference_model
import torch
import torch.nn as nn


def custom_batch_norm(input, channels, with_mean):
    eps = 0.00001
    out = torch.zeros_like(input)
    for i in range(channels):
        if with_mean:
            mu = torch.mean(input, dim=1)
        else:
            mu = 0
        var = torch.sqrt(torch.var(input, dim=1)+eps)
        out[:,i,...] = (input[:, i, ...] - mu)/var
    return out


class BasicBlock(nn.Module):
    expansion = 1
    def __init__(self, in_layer, out_layer, stride=1, downsample=None, use_bias=True):
        super(BasicBlock, self).__init__()
        self.conv1 = nn.Conv2d(in_layer, out_layer, kernel_size=3, stride=stride, padding=1, bias=use_bias)
        self.bn1 = nn.BatchNorm2d(out_layer, track_running_stats=False)
        self.relu = nn.ReLU(inplace=True)
        self.conv2 = nn.Conv2d(out_layer, out_layer, kernel_size=3, stride=stride, padding=1, bias=use_bias)
        self.bn2 = nn.BatchNorm2d(out_layer, track_running_stats=False)
        self.downsample = downsample
        self.stride = stride

    def forward(self, x):
        residual = x
        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)
        out = self.conv2(out)
        out = self.bn2(out)
        if self.downsample is not None:
            residual = self.downsample(x)
        out += residual
        out = self.relu(out)
        return out

    
class Resnet(base_inference_model.InferenceModel, nn.Module):
    """
        Class Implementing the ResNet model.
        Methods:
            - setOpts
                inputs: a Dict containing the key and value for a new configuration setting
            - forward
                inputs: signal (measured signal);
                outputs: Estimated parameters
    """

    def __init__(self, config_object):
        super(Resnet, self).__init__()
        self.__name__ = 'RESNET'
        self.__require_initial_guess__ = False
        self.use_bias = config_object.args.inference.useBias
        self.args = config_object.args
        self.__buildNetwork__()

    def __buildNetwork__(self):
        self.input_planes = self.args.inference.outputChannelsLayer1
        self.conv1 = nn.Conv2d(self.args.inference.inputChannels, self.args.inference.outputChannelsLayer1, kernel_size=1, stride=1, padding=0, bias=self.use_bias)
        self.bn1 = nn.BatchNorm2d(self.args.inference.outputChannelsLayer1, track_running_stats=False)
        self.relu = nn.ReLU(inplace=True)
        
        self.layer1 = self.__makeLayer__(BasicBlock, self.args.inference.outputChannelsLayer2, self.args.inference.convLayersInResidualBlock)
        self.layer2 = self.__makeLayer__(BasicBlock, self.args.inference.outputChannelsLayer3, self.args.inference.convLayersInResidualBlock)
        self.layer3 = self.__makeLayer__(BasicBlock, self.args.inference.outputChannelsLayer4, self.args.inference.convLayersInResidualBlock)
        self.layer4 = self.__makeLayer__(BasicBlock, self.args.inference.outputChannelsLayer5, self.args.inference.convLayersInResidualBlock)
        self.layer5 = self.__makeLayer__(BasicBlock, self.args.inference.outputChannelsLayer6, self.args.inference.convLayersInResidualBlock)
        self.layer6 = self.__makeLayer__(BasicBlock, self.args.inference.outputChannelsLayer7, self.args.inference.convLayersInResidualBlock)
        self.layer_out = nn.Conv2d(self.args.inference.outputChannelsLayer7, self.args.inference.outputChannels, kernel_size=1, stride=1, padding=0, bias=self.use_bias)
        self.use_mean = self.args.inference.useMeanBatchNorm

        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                n = m.kernel_size[0] * m.kernel_size[1] * m.out_channels
                torch.nn.init.kaiming_normal_(m.weight)
            elif isinstance(m, nn.BatchNorm2d):
                m.weight.data.fill_(1)
                m.bias.data.zero_()

    def __makeLayer__(self, block, channels, blocks, stride=1):
        downsample = None
        if stride != 1 or self.input_planes != channels*block.expansion:
            downsample = nn.Sequential(nn.Conv2d(self.input_planes, channels*block.expansion,
                                                 kernel_size=1, stride=stride, bias=self.use_bias),
                                       nn.BatchNorm2d(channels*block.expansion, track_running_stats=False))
        layers = []
        layers.append(block(self.input_planes, channels, stride, downsample, self.use_bias))
        self.input_planes = channels*block.expansion
        for _ in range(1, blocks):
            layers.append(block(self.input_planes, channels))
        return nn.Sequential(*layers)

    def forward(self, inputs):
        signal = inputs
        x = self.conv1(signal)
        x = self.bn1(x)
        # x = custom_batch_norm(x, self.args.inference.inputChannels, self.use_mean)
        x = self.relu(x)
        x = self.layer1(x)
        x = self.layer2(x)
        x = self.layer3(x)
        x = self.layer4(x)
        x = self.layer5(x)
        x = self.layer6(x)
        out = self.layer_out(x)
        return out



from os.path import dirname, realpath
filepath = realpath(__file__)
parent_dir = dirname(filepath)
rec = 1 # This is always related to where the module is compared to the root of EMCqMRI. In this case, 4 folders down.
for _ in range(rec):
    parent_dir = dirname(parent_dir)
import sys
sys.path.insert(0,parent_dir)

from .base import *
from .engine import build_model, data_loader, estimate, train_model
# from .configuration import pll_configuration
# from .configuration.engine import *
# from .configuration.inference_model import *
from .utilities import *



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from abc import ABC, abstractmethod
import torch

class Likelihood(ABC):
    """Base class for implementation of likelihood models
    """

    def __init__(self, config_object, ll_obj):
        """
        Args:
            config_object ([Configuration]): Configuration object where following attributes
                must be specified:

                - args.engine.signal_model ([SignalModel])

            ll_obj ([Likelihood]): Circular reference to child of Likelihood object
        """

        super().__init__()
        self.args = config_object.args
        self.ll_model = ll_obj

    @abstractmethod
    def likelihood(self, signal, modeled_signal):
        """
        Computes the loss, or error, based on the negative log likelihood function.

        Args:
            signal ([torch.Tensor]): Measured, input signal.

            modeled_signal ([torch.Tensor]): Tensor containing a simulated signal, generated
                with a signal model.
             
        Raises:
            NotImplementedError: When the subclass does not override this method.

        Returns:
            ([torch.Float]): A scalar loss (i.e. error)
        """
        raise NotImplementedError("Likelihood Function not implemented")

    def gradients(self, signal, kappa, *extra_args):
        """
        Computes the gradient of the signal model parameters with respect to the likelihood function.
        This function can be overriden if you want to define your own gradients (e.g. analytical, 
        different shapes, etc.)

        Args:
            signal ([torch.Tensor]): Measured, input signal.
            kappa ([list]): A list of torch.Tensor parameters.
            *extra_args ([tuple]): Any additional parameters required by the signal model or 
                likelihood model
             
        Raises:
            TypeError: When kappa is not a list of torch.Tensor.

        Returns:
            ([list]): list of torch.Tensor with same number of elements as Kappa. Each element of the list
            is the gradient of each parameter with respect to the likelihood function.
        """
        with torch.enable_grad():
            weighted_images = self.args.engine.signal_model.forward(kappa, *extra_args)
            loss = self.ll_model.likelihood(signal, weighted_images)
            loss.backward()
            if isinstance(kappa,list):
                if isinstance(kappa[0],list): # This is for when kappa contains more than 1 type of parameter
                    param_map_gradient = []
                    for kappa_map in kappa:
                        gradient = ([param_map.grad for param_map in kappa_map])
                        param_map_gradient.append(torch.stack(gradient))
                else:
                    param_map_gradient = torch.stack([param_map.grad for param_map in kappa])
            else:
                param_map_gradient = kappa.grad
        return param_map_gradient


        
